/*
   Siag, Scheme In A Grid
   Copyright (C) 2000  Ulric Eriksson <ulric@siag.nu>
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.
 */

#include "../config.h"

#ifdef HAVE_LIBCLNTSH

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/time.h>
#include <unistd.h>
#include <oratypes.h>
#include <ociapr.h>
#include <ocidem.h>

#include "../siod/siod.h"
#include "calc.h"

#define MAX_ITEM_BUFFER_SIZE    128
#define MAX_SELECT_LIST_SIZE    64
#define PARSE_DEFERRED		0
#define V7_BEHAVIOR		2
#define NUMBER_WIDTH		8

/*
 * structures
 */

typedef struct
{
	sb4 dbsize;
	sb2 dbtype;
	sb1 buf[MAX_ITEM_BUFFER_SIZE];
	sb4 buflen;
	sb4 dsize;
	sb2 prec;
	sb2 scale;
	sb2 nullok;
} DescribeStruct;

typedef struct
{
	ub1 buf[MAX_ITEM_BUFFER_SIZE];
	float flt_buf;
	sword int_buf;
	sb2 indp;
	ub2 col_retlen;
	ub2 col_retcode;
} DefineStruct;

static void handle_error (Lda_Def *lda, Cda_Def *cda)
{
    text msg[512];
    sword n;

    n = oerhms(lda, cda->rc, msg, (sword) sizeof (msg));
    fprintf (stderr, "%.*s\n", n, msg);
}

static LISP store_result(buffer *buf, int sht, int row0, int col0,
		Lda_Def *lda, Cda_Def *cda, DescribeStruct *desc,
		DefineStruct *def, sword ncols)
{
	sword col;
	long nrows = 0;
	cval val;
	LISP res = NIL;
	char b[100];
	char *p;

	val.number = 0;
	while (1) {
		if (ofetch (cda)) {
			if (cda->rc == NO_DATA_FOUND) {
				break;
			}
			if (cda->rc != NULL_VALUE_RETURNED) {
				handle_error (lda, cda);
				break;
			}	
	        }

		for (col = 0; col < ncols ; col++) {
			switch (desc[col].dbtype) {
			case FLOAT_TYPE:
				sprintf(b, "%0.*f", desc[col].scale,
					def[col].flt_buf);
				p = b;
				break;
			case INT_TYPE:
				sprintf(b, "%ld", (long)def[col].int_buf);
				p = b;
				break;
			default:
				p = def[col].buf;
				break;
			}
			if (res == NIL) {
				res = strcons(strlen(p), p);
			} else {
				ins_data(buf, siod_interpreter,
					p, val, MTEXT,
					sht, row0+nrows, col0+col);
			}
		}
		++nrows;
	}

	return res;
}

static sword describe_define (Lda_Def *lda, Cda_Def *cda, DescribeStruct *desc,
			DefineStruct *def)
{
	sword col, deflen, deftyp;
	static ub1 *defptr;

	for (col = 0; col < MAX_SELECT_LIST_SIZE; col++) {
		desc[col].buflen = MAX_ITEM_BUFFER_SIZE;
		if (odescr (cda, col + 1, &desc[col].dbsize,
			&desc[col].dbtype, &desc[col].buf[0],
			&desc[col].buflen, &desc[col].dsize,
			&desc[col].prec, &desc[col].scale,
			&desc[col].nullok)) {
			if (cda->rc == VAR_NOT_IN_LIST) {
				break;
			} else {
				handle_error (lda, cda);
				return 1;
			}
		}

		switch (desc[col].dbtype) {
		case NUMBER_TYPE:
			desc[col].dbsize = NUMBER_WIDTH;
			if (desc[col].scale != 0) {
				defptr = (ub1 *) &def[col].flt_buf;
				deflen = (sword) sizeof(float);
				deftyp = FLOAT_TYPE;
				desc[col].dbtype = FLOAT_TYPE;
			} else {
				defptr = (ub1 *) &def[col].int_buf;
				deflen = (sword) sizeof(sword);
				deftyp = INT_TYPE;
				desc[col].dbtype = INT_TYPE;
			}
			break;
		default:
			if (desc[col].dbtype == DATE_TYPE) {
				desc[col].dbsize = 32;
			}
			if (desc[col].dbtype == ROWID_TYPE) {
				desc[col].dbsize = 18;
			}
			defptr = def[col].buf;
			deflen = desc[col].dbsize > MAX_ITEM_BUFFER_SIZE ?
				MAX_ITEM_BUFFER_SIZE : desc[col].dbsize + 1;
			deftyp = STRING_TYPE;
			break;
		}

		if (odefin(cda, col + 1, defptr, deflen, deftyp, -1,
			&def[col].indp, (text *) 0, -1, -1,
			&def[col].col_retlen, &def[col].col_retcode)) {
			handle_error(lda, cda);
			return 1;
		}
	}

	return col;
}

static LISP exec_sql (buffer *buf, int sht, int row0, int col0,
		char *query, Lda_Def *lda, Cda_Def *cda,
		DescribeStruct *desc, DefineStruct *def)
{
	sword ncols = 0, sql_function;
	LISP res = NIL;

	if (oparse (cda, (text *) query, (sb4) -1,
		(sword) PARSE_DEFERRED, (ub4) V7_BEHAVIOR)) {
		handle_error(lda, cda);
		return NIL;
	}

	sql_function = cda->ft;

	if (sql_function == FT_SELECT) {
		if ((ncols = describe_define (lda, cda, desc, def)) == -1) {
			return NIL;
		}
	}

	if (oexec (cda)) {
		handle_error(lda, cda);
		return NIL;
	}

	if (sql_function == FT_SELECT) {
		res = store_result(buf, sht, row0, col0,
				 lda, cda, desc, def, ncols);
	}

	return res;
}

static LISP loracle_query(LISP query)
{
	char *username = strdup("admin"), *password = strdup("admin");
	DescribeStruct desc[MAX_SELECT_LIST_SIZE];
	DefineStruct def[MAX_SELECT_LIST_SIZE];
	char *q = get_c_string(query);
	LISP res;
	int row0, col0, sht;
	buffer *buf;
	Lda_Def lda;
	Cda_Def cda;
	ub4 hda[HDA_SIZE/(sizeof(ub4))];

	get_siod_coords(&row0, &col0, &sht, &buf);

	memset(&lda, 0, sizeof lda);
	memset(&cda, 0, sizeof cda);
	memset(hda, 0, sizeof hda);

	if (olog (&lda, (ub1 *) hda, (text *)username, -1,
		(text *)password, -1,
		(text *) 0, -1, (ub4) OCI_LM_DEF)) {
			printf ("Unable to connect to ORACLE\n");
			return NIL;
	}
 
	if (oopen (&cda, &lda, (text *) 0, -1, -1, (text *) 0, -1)) {
		fprintf (stderr, "Cannot open cursor\n");
		ologof (&lda);
		return NIL;
        }

	res = exec_sql(buf, sht, row0, col0, q, &lda, &cda, desc, def);

	if (oclose (&cda)) {
		fprintf (stderr, "Error closing cursor\n");
	}
	if (ologof (&lda)) {
		fprintf (stderr, "Error logging off\n");
	}
 
        return res;
}

void init_oracle(void)
{
	init_subr_1("oracle_query", loracle_query);
}

#else

void init_oracle(void)
{
	;
}

#endif	/* ORACLE */

